/*
 * Emulate a basic 6502 (MOS) chip.
 */

var Memory = require('src/memory');

class CPU{
  constructor(){
    // Registers
    this.__PC = 0;   // Program Counter (16 bit)
    this.__IRQ = 0;  // IRQ interrupt address code (16 bit)
    this.__SR = 0;   // Status Register (8 bit)
    this.__XR = 0;   // X Register (8 bit)
    this.__YR = 0;   // Y Register (8 bit)
    this.__AR = 0;   // Accumulator Register (8 bit)

    // Variables to watch for Interrupts.
    this.__nmi = false;
    this.__irq = false;

    // Memory module or controller.
    this.__mem = null; // Must be explicitly attached.

    // Hold any created CLK instances.
    this.__clkfn = null;
  }

  set NMI(n){
    this.__nmi = (n === true);
  }

  set IRQ(q){
    // TODO: Verify this.
    // TODO: Do not set if the interrupt flag is off.
    this.__irq = (q === true);
  }

  reset(){
    // TODO: Read memory address $FFFC - FFFD to find PC start location.
    // TODO: Reset status registers that get changed on a reset.
  }

  clk(){
    if (this.__clkfn === null){
      this.__clkfn = (function(){
        // TODO: All the work!!
      }).bind(this);
    }
    return this.__clkfn;
  }

  memory(mem){
    if (!(mem instanceof Memory))
      throw new ValueError("Expected Memory instance object.");
    this.__mem = mem;
  }
}



// ---------------------------------------------------------------------
// ---------------------------------------------------------------------

function toHexString(v, l){
  l = (typeof(l) === 'number' && (l === 8 || l === 16)) ? l : 8;
  var r = v.toString(16);
  if (r.length < 2)
    r = "0" + r;
  if (l === 16 && r.length < 4)
    r = ((r.length === 2) ? "00" : "0") + r;
  return r;
}

function tokenize(l){
  var m = l.match(/([^()]+)|([()])/g);
  var ip = false;
  t = [];
  m.forEach((mi)=>{
    if (mi === "("){
      ip = true;
    } else {
      if (ip){
        mi.split(/[\s,]+/).forEach((i)=>{
          if (i !== "")
            t.push("+" + i);
        });
      } else {
        mi.split(/[\s,]+/).forEach((i)=>{
          if (i !== "")
            t.push(i);
        });
      }
    }
  });
}

class Assembler{
  constructor(initpc){
    // Labels that hold variable values.
    this.__varlabels = {};
    // Labels that hold jump/branch locations.
    this.__jmplabels = {};
    // Program counter to track where in the code/memory a
    // operation is located.
    this.__PC = (initpc >= 0) ? initpc : 0;
  }

  function __AddrModeVal(tA, tB){
    var mode = "";
    var v = Number.NaN;
    if (tA[0] === '#'){
      mode = "i"; // Immediate
      if (tA[1] === '$'){
        v = parseInt(tA.substr(2), 16);
      } else if (tA[1] === 'b'){
        v = parseInt(tA.substr(2), 2);
      } else {
        v = parseInt(tA.substr(1));
        if (isNaN(v)){
          let lbl = tA.substr(1);
          if (lbl in this.__varlabels){
            if (this.__varlabels[lbl] < 256)
              v = this.__varlabels[lbl]
          }
        }
      }
    } else if (tA[0] === '$'){
      v = parseInt(tA.substr(1), 16);
      if (tA.length === 3){
        // Zero Page
        if (tB !== null){
          if (tB === "X")
            mode = "zX";
        } else {
          mode = "z";
        }
      } else if (tA.length === 5){
        // Absolute
        if (tB !== null){
          if (tB === "X"){
            mode = "aX";
          } else if (tB === "Y"){
            mode = "aY";
          }
        } else {
          mode = "a";
        }
      }
    } else if (tA.startsWith("+")){
      if (tB !== null){
        if (tB === "+X"){ // The plus exists due to how tokenize() works.
          mode = "nX";
        } else if (tB === "Y"){
          mode = "nY";
        }

        if (mode !== ""){
          if (tA.startsWith("+$")){
            v = parseInt(tA.substr(2));
          } else {
            let lbl = tA.substr(1);
            if (lbl in this.__varlabels){
              if (this.__varlabels[lbl] < 256)
                v = vlb[lbl];
            }
          }
        }
      }
    } else {
      // We've been given a variable name. Need to look it up.
      // if the var is 1 byte, assume "Zero Page". If var 2 bytes, assume "Absolute"
      if (tA in this.__varlabel){
        v = this.__varlabel[tA];
        if (v < 256){
          if (tB !== null){
            if (tB === "X")
              mode = "zX";
          } else {
            mode = "z";
          }
        } else {
          if (tB !== null){
            if (tB === "X"){
              mode = "aX";
            } else if (tB === "Y"){
              mode = "aY";
            }
          } else {
            mode = "a";
          }
        }
      }
    }

    if (mode !== "" && ! isNaN(v))
      return [mode, v];
    throw new Error("Malformed op-code or value on program address " + toHexString(this.__PC));
  }

  compile(src){
    var op = [];
    src.split("\n").forEach((line)=>{
      line = line.trim();
      if (line[0] !== ";"){ // Skip comment lines.
        line = line.split(";")[0].trim(); // Take out any trailing comments.
        var tokens = tokenize(line);

        if (tokens[0] === 'define'){
          // Variable label!!
        } else if (tokens[0][tokens[0].length - 1] === ':'){
          // Jump label!!
        } else if (tokens[0].length === 3){

          let StoreSingleOp = (code)=>{
            if (tokens.length === 1){
              op.push(code);
              this.__PC += 1;
              return true;
            }
            return false;
          };

          let StoreBranchOp = (code) => {
            if (tokens.length === 2){
              if (token[1][0] === '$' && (tokens[1].length === 3 || tokens[1].length === 5)){
                let v = parseInt(token[1].substr(1), 16);
                if (isNaN(v))
                  return false;
                if (v < this.__PC - 127 || v > this.__PC + 128)
                  throw new Error("Branch exceeds maximum number of bytes on program address " + toHexString(this.__PC));
                v = v - this.__PC;
                op.push(code);
                op.push((v >= 0) ? v : v + 255);
                this.__PC += 2;
              }
            }
            return false;
          };

          let StoreOp = (codes, mint, maxt) => {
            if (tokens.length >= mint && tokens.length <= maxt){
              let mv = addrModeVal(tokens[1], (tokens.length === 3) ? tokens[2].toUpperCase(), null);
              let modea = true;
              switch(mv[0]){
                case "i":
                  modea = codes[0] !== null;
                  if (modea){
                    op.push(codes[0]);
                    op.push(mv[1]);
                    this.__PC += 2;
                  } break;
                case "z":
                  modea = codes[1] !== null;
                  if (modea){
                    op.push(codes[1]);
                    op.push(mv[1]);
                    this.__PC += 2;
                  } break;
                case "zX":
                  modea = codes[2] !== null;
                  if (modea){
                    op.push(codes[2]);
                    op.push(mv[1]);
                    this.__PC += 2;
                  } break;
                case "a":
                  modea = codes[3] !== null;
                  if (modea){
                    op.push(codes[3]);
                    op.push(mv[1] & 0x000000FF);
                    op.push((mv[1] & 0x0000FF00) >> 8);
                    this.__PC += 3;
                  } break;
                case "aX":
                  modea = codes[4] !== null;
                  if (modea){
                    op.push(codes[4]);
                    op.push(mv[1] & 0x000000FF);
                    op.push((mv[1] & 0x0000FF00) >> 8);
                    this.__PC += 3;
                  } break;
                case "aY":
                  modea = codes[5] !== null;
                  if (modea){
                    op.push(codes[5]);
                    op.push(mv[1] & 0x000000FF);
                    op.push((mv[1] & 0x0000FF00) >> 8);
                    this.__PC += 3;
                  } break;
                case "nX":
                  modea = codes[6] !== null;
                  if (modea){
                    op.push(codes[6]);
                    op.push(mv[1]);
                    this.__PC += 2;
                  } break;
                case "nY":
                  modea = codes[7] !== null;
                  if (modea){
                    op.push(codes[7]);
                    op.push(mv[1]);
                    this.__PC += 2;
                  } break;
              }
              if (modea === false)
                throw new Error("Op-code does not support implied mode on program address " + toHexString(this.__PC));
              return true;
            }
            return false;
          };

          let procFailed = false;
          let mv = null;
          // Possible op code.
          switch(tokens[0].toLowerCase()){
// --- ADC
            case 'adc':
              procFailed = StoreOp([0x69, 0x65, 0x75, 0x6D, 0x7D, 0x79, 0x61, 0x71], 2, 3);
              /*
              if (tokens.length >= 2 && tokens.length <= 3){
                mv = addrModeVal(tokens[1], (tokens.length === 3) ? tokens[2].toUpperCase(), null);
                this.__PC += 2;
                switch(mv[0]){
                  case "i":
                    op.push(0x69);
                    op.push(mv[1]);
                    break;
                  case "z":
                  case "zX":
                    op.push((mv[0] === "z") ? 0x65 : 0x75);
                    op.push(mv[1]);
                    break;
                  case "a":
                  case "aX":
                  case "aY":
                    op.push((mv[0] === "a") ? 0x6D : ((mv[0] === "aX") ? 0x7D : 0x79));
                    op.push(mv[1] & 0x000000FF);
                    op.push((mv[1] & 0x0000FF00) >> 8);
                    this.__PC += 1;
                    break;
                  case "nX":
                  case "nY":
                    op.push((mv[0] === "nX") ? 0x61 : 0x71);
                    op.push(mv[1]);
                    break;
                }
              } else { procFailed = true; }
              */
              break;
// --- AND
            case 'and':
              procFailed = StoreOp([0x29, 0x25, 0x35, 0x2D, 0x3D, 0x39, 0x21, 0x31], 2, 3); 
              break;
// --- ASL
            case 'asl':
              if (tokens.length === 2){
                if (tokens[1].toUpperCase() === 'A'){
                  op.push(0x0A);
                  this.__PC += 1;
                  break;
                }
              }
              procFailed = StoreOp([null, 0x06, 0x16, 0x0E, 0x1E, null, null, null], 2, 3);  
              break;
// --- BCC
            case 'bcc':
              procFailed = StoreBranchOp(0x90); break;
// --- BCS
            case 'bcs':
              procFailed = StoreBranchOp(0xB0); break;
// --- BEQ
            case 'beq':
              procFailed = StoreBranchOp(0xF0); break;
// --- BIT
            case 'bit':
              procFailed = StoreOp([null, 0x24, null, 0x2C, null, null, null, null], 2, 2); break;
// --- BMI
            case 'bmi':
              procFailed = StoreBranchOp(0x30); break;
// --- BNE
            case 'bne':
              procFailed = StoreBranchOp(0xD0); break;
// --- BPL
            case 'bpl':
              procFailed = StoreBranchOp(0x10); break;
// --- BRK
            case 'brk':
              procFailed = StoreSingleOp(0x00); break;
// --- BVC
            case 'bvc':
              procFailed = StoreBranchOp(0x50); break;
// --- BVS
            case 'bvs':
              procFailed = StoreBranchOp(0x70); break;
// --- CLC
            case 'clc':
              procFailed = StoreSingleOp(0x18); break;
// --- CLD
            case 'cld':
              procFailed = StoreSingleOp(0xD8); break;
// --- CLI
            case 'cli':
              procFailed = StoreSingleOp(0x58); break;
// --- CLV
            case 'clv':
              procFailed = StoreSingleOp(0xB8); break;
// --- CMP
            case 'cmp':
              procFailed = StoreOp([0xC9, 0xC5, 0xD5, 0xCD, 0xDD, 0xD9, 0xC1, 0xD1], 2, 3); break;
// --- CPX
            case 'cpx':
              procFailed = StoreOp([0xE0, 0xE4, null, 0xEC, null, null, null, null], 2, 2); break;
// --- CPY
            case 'cpy':
              procFailed = StoreOp([0xC0, 0xC4, null, 0xCC, null, null, null, null], 2, 2); break; 
// --- DEC
            case 'dec':
              procFailed = StoreOp([null, 0xC6, 0xD6, 0xCE, 0xDE, null, null, null], 2, 3); break; 
// --- DEX
            case 'dex':
              procFailed = StoreSingleOp(0xCA); break;
// --- DEY
            case 'dey':
              procFailed = StoreSingleOp(0x88); break;
// --- EOR
            case 'eor':
              procFailed = StoreOp([0x49, 0x45, 0x55, 0x4D, 0x5D, 0x59, 0x41, 0x51], 2, 3); break; 
// --- INC
            case 'inc':
              procFailed = StoreOp([null, 0xE6, 0xF6, 0xEE, 0xFE, null, null, null], 2, 3); break; 
// --- INX
            case 'inx':
              procFailed = StoreSingleOp(0xE8); break;
// --- INY
            case 'iny':
              procFailed = StoreSingleOp(0xC8); break;
// --- JMP
            case 'jmp':
              if (tokens.length === 2){
                if (tokens[1].startsWith("$") && tokens[1].length === 5){
                  let v = parseInt(tokens[1].substr(1), 16);
                  if (!isNaN(v)){
                    op.push(0x4C);
                    op.push(v & 0x000000FF);
                    op.push((v & 0x0000FF00) >> 8);
                    this.__PC += 3;
                  } else {
                    throw new Error("Malformed op-code or value on program address " + toHexString(this.__PC));
                  }
                } else if (tokens[1].startsWith("+$") && tokens[1].length === 6){
                  let v = parseInt(tokens[1].substr(2), 16);
                  if (!isNaN(v)){
                    op.push(0x6C);
                    op.push(v & 0x000000FF);
                    op.push((v & 0x0000FF00) >> 8);
                    this.__PC += 3;
                  } else {
                    throw new Error("Malformed op-code or value on program address " + toHexString(this.__PC));
                  }
                } else { procFailed = true; }
              } else { procFailed = true; }
              break;
// --- JSR
            case 'jsr':
              if (tokens.length === 2){
                if (tokens[1].startsWith("$") && tokens[1].length === 5){
                  let v = parseInt(tokens[1].substr(1), 16);
                  if (!isNaN(v)){
                    op.push(0x20);
                    op.push(v & 0x000000FF);
                    op.push((v & 0x0000FF00) >> 8);
                    this.__PC += 3;
                  } else {
                    throw new Error("Malformed op-code or value on program address " + toHexString(this.__PC));
                  }
                } else { procFailed = true; }
              } else {
                procFailed = true;
              }
              break;
// --- LDA
            case 'lda':
              procFailed = StoreOp([0xA9, 0xA5, 0xB5, 0xAD, 0xBD, 0xB9, 0xA1, 0xB1], 2, 3); break; 
              break;
// --- LDX
            case 'ldx':
              procFailed = StoreOp([0xA2, 0xA6, 0xB6, 0xAE, 0xBE, null, null, null], 2, 3); break;
              break;
// --- LDY
            case 'ldy':
              procFailed = StoreOp([0xA0, 0xA4, 0xB4, 0xAC, 0xBC, null, null, null], 2, 3); break;
              break;
// --- LSR
            case 'lsr':
              if (tokens.length === 2){
                if (tokens[1].toUpperCase() === 'A'){
                  op.push(0x4A);
                  this.__PC += 1;
                  break;
                }
              }
              procFailed = StoreOp([null, 0x46, 0x56, 0x4E, 0x5E, null, null, null], 2, 3); break;
// --- NOP
            case 'nop':
              procFailed = StoreSingleOp(0xEA); break;
// --- ORA
            case 'ora':
              procFailed = StoreOp([0x09, 0x05, 0x15, 0x0D, 0x1D, 0x19, 0x01, 0x11], 2, 3); break;
// --- PHA
            case 'pha':
              procFailed = StoreSingleOp(0x48); break;
// --- PHP
            case 'php':
              procFailed = StoreSingleOp(0x08); break;
// --- PLA
            case 'pla':
              procFailed = StoreSingleOp(0x68); break;
// --- PLP
            case 'plp':
              procFailed = StoreSingleOp(0x28); break;
// --- ROL
            case 'rol':
              if (tokens.length === 2){
                if (tokens[1].toUpperCase() === 'A'){
                  op.push(0x2A);
                  this.__PC += 1;
                  break;
                }
              }
              procFailed = StoreOp([null, 0x26, 0x36, 0x2E, 0x3E, null, null, null], 2, 3); break;
// --- ROR
            case 'ror':
              if (tokens.length === 2){
                if (tokens[1].toUpperCase() === 'A'){
                  op.push(0x6A);
                  this.__PC += 1;
                  break;
                }
              }
              procFailed = StoreOp([null, 0x66, 0x76, 0x6E, 0x7E, null, null, null], 2, 3); break;
// --- RTI
            case 'rti':
              procFailed = StoreSingleOp(0x40); break;
// --- RTS
            case 'rts':
              procFailed = StoreSingleOp(0x60); break;
              break;
// --- SBC
            case 'sbc':
              procFailed = StoreOp([0xE9, 0xE5, 0xF5, 0xED, 0xFD, 0xF9, 0xE1, 0xF1], 2, 3); break;
// --- SEC
            case 'sec':
              procFailed = StoreSingleOp(0x38); break;
// --- SED
            case 'sed':
              procFailed = StoreSingleOp(0xF8); break;
// --- SEI
            case 'sei':
              procFailed = StoreSingleOp(0x78); break;
              break;
// --- STA
            case 'sta':
              procFailed = StoreOp([null, 0x85, 0x95, 0x8D, 0x9D, 0x99, 0x81, 0x91], 2, 3); break;
// --- STX
            case 'stx':
              procFailed = StoreOp([null, 0x86, 0x96, 0x8E, null, null, null, null], 2, 3); break;
// --- STY
            case 'sty':
              procFailed = StoreOp([null, 0x84, 0x94, 0x8C, null, null, null, null], 2, 3); break;
// --- TAX
            case 'tax':
              procFailed = StoreSingleOp(0xAA); break;
// --- TAY
            case 'tay':
              procFailed = StoreSingleOp(0xA8); break;
// --- TSX
            case 'tsx':
              procFailed = StoreSingleOp(0xBA); break;
// --- TXA
            case 'txa':
              procFailed = StoreSingleOp(0x8A); break;
// --- TXS
            case 'txs':
              procFailed = StoreSingleOp(0x9A); break;
// --- TYA
            case 'tya':
              procFailed = StoreSingleOp(0x98); break;
// --- ---
            default:
              throw new Error("Unknown op-code '" + tokens[0].toUpperCase() + "' at program address " + toHexString(this.__PC));
          }

          if (procFailed)
            throw new Error("Malformed op-code on program address " + toHexString(this.__PC));
        } else {
          throw new Error("Failed to compile line '" + line + "' at program address " + toHexString(this.__PC));
        }
      }
    });
    return new Uint8Array(op);
  }
}



// ---------------------------------------------------------------------
// ---------------------------------------------------------------------

module.exports = {
  CPU: CPU,
  Assembler: Assembler
};




